//-----------------------------------------------------------------------------
// Copyright (c) 2006-2007 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
//
// This is the first demo in the Direct3D 9 HLSL (High Level Shading Language)
// demo series. This demo implements the per-vertex Blinn-Phong lighting model
// that fixed function Direct3D uses with a series of HLSL shaders.
//
// The purpose of this demo is to replicate the fixed function lighting model.
// Switching between the fixed function pipeline and the programmable pipeline
// should show no visible differences within the scene in this demo. The
// shaders presented in this demo will serve as the foundation for future
// Direct3D 9 HLSL demos.
//
// A single Direct3D Effect File (.fx) is provided containing 3 techniques.
// There is a technique for each of the three fixed function lighting models:
// directional lighting, point lighting, and spotlight lighting.
//
// The demo consists of a single textured cube centered at the world origin.
// The light source is fixed and located directly in front of the cube where
// the virtual camera is located in the scene. The mouse can be used to rotate
// the cube and to zoom in and out of the scene. Note that the cube
// intentionally has a very low polygon count to emphasize the fact that
// realistic per-vertex lighting depends on sufficiently tessellated geometry.
//
//-----------------------------------------------------------------------------

#if !defined(WIN32_LEAN_AND_MEAN)
#define WIN32_LEAN_AND_MEAN
#endif

#if defined(_DEBUG)
#define D3D_DEBUG_INFO
#endif

#include <windows.h>
#include <d3d9.h>
#include <d3dx9.h>
#include <sstream>
#include <stdexcept>
#include <string>

#if defined(_DEBUG)
#include <crtdbg.h>
#endif

//-----------------------------------------------------------------------------
// Macros.
//-----------------------------------------------------------------------------

#define SAFE_RELEASE(x) if ((x) != 0) { (x)->Release(); (x) = 0; }

//-----------------------------------------------------------------------------
// Constants.
//-----------------------------------------------------------------------------

#if !defined(CLEARTYPE_QUALITY)
#define CLEARTYPE_QUALITY 5
#endif

#if !defined(WHEEL_DELTA)
#define WHEEL_DELTA 120
#endif

#if !defined(WM_MOUSEWHEEL)
#define WM_MOUSEWHEEL 0x020A
#endif

#define APP_TITLE "D3D9 HLSL Vertex Lighting Demo"

const float CAMERA_FOVY = D3DXToRadian(45.0f);
const float CAMERA_ZNEAR = 0.01f;
const float CAMERA_ZFAR = 100.0f;

const float DOLLY_MIN = -10.0f;
const float DOLLY_MAX = -2.5f;

const float LIGHT_SPOT_INNER_CONE = D3DXToRadian(30.0f);
const float LIGHT_SPOT_OUTER_CONE = D3DXToRadian(50.0f);

const float MOUSE_ORBIT_SPEED = 0.30f;
const float MOUSE_DOLLY_SPEED = 0.02f;
const float MOUSE_TRACK_SPEED = 0.005f;
const float MOUSE_WHEEL_DOLLY_SPEED = 0.005f;

//-----------------------------------------------------------------------------
// Types.
//-----------------------------------------------------------------------------

enum {DIR_LIGHT, POINT_LIGHT, SPOT_LIGHT};

struct Vertex
{
    float pos[3];
    float texCoord[2];
    float normal[3];
};

//-----------------------------------------------------------------------------
// Globals.
//-----------------------------------------------------------------------------

HWND                         g_hWnd;
HINSTANCE                    g_hInstance;
D3DPRESENT_PARAMETERS        g_params;
IDirect3D9                  *g_pDirect3D;
IDirect3DDevice9            *g_pDevice;
ID3DXFont                   *g_pFont;
ID3DXEffect                 *g_pEffect;
IDirect3DVertexDeclaration9 *g_pVertexDecl;
IDirect3DVertexBuffer9      *g_pVertexBuffer;
IDirect3DTexture9           *g_pColorMapTexture;
IDirect3DTexture9           *g_pNullTexture;
bool                         g_enableVerticalSync;
bool                         g_isFullScreen;
bool                         g_hasFocus;
bool                         g_displayHelp;
bool                         g_disableColorMapTexture;
bool                         g_fixedFunction;
DWORD                        g_msaaSamples;
DWORD                        g_maxAnisotrophy;
int                          g_framesPerSecond;
int                          g_windowWidth;
int                          g_windowHeight;
int                          g_lightType = DIR_LIGHT;
float                        g_heading;
float                        g_pitch;
float                        g_sceneAmbient[4] = {0.2f, 0.2f, 0.2f, 1.0f};
D3DXVECTOR3                  g_cameraPos(0.0f, 0.0f, -5.0f);
D3DXVECTOR3                  g_targetPos(0.0f, 0.0f, 0.0f);

D3DLIGHT9 g_light =
{
    D3DLIGHT_DIRECTIONAL,           // type
    1.0f, 1.0f, 1.0f, 0.0f,         // diffuse
    1.0f, 1.0f, 1.0f, 0.0f,         // specular
    1.0f, 1.0f, 1.0f, 0.0f,         // ambient
    0.0f, 0.0f, 0.0f,               // position
    0.0f, 0.0f, 1.0f,               // direction
    1000.0f,                        // range
    1.0f,                           // falloff
    1.0f,                           // attenuation0 (constant)
    0.0f,                           // attenuation1 (linear)
    0.0f,                           // attenuation2 (quadratic)
    0.0f,                           // theta
    0.0f                            // phi
};

D3DMATERIAL9 g_material =
{
    0.8f, 0.8f, 0.8f, 1.0f,         // diffuse
    0.2f, 0.2f, 0.2f, 1.0f,         // ambient
    0.0f, 0.0f, 0.0f, 1.0f,         // specular
    0.0f, 0.0f, 0.0f, 1.0f,         // emissive
    0.0f                            // power
};

D3DVERTEXELEMENT9 g_vertexElements[] =
{
    {0,  0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
    {0, 12, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
    {0, 20, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,   0},
    D3DDECL_END()
};

Vertex g_cube[36] =
{
    // -Z face
    {-1.0f,  1.0f, -1.0f,   0.0f, 0.0f,    0.0f,  0.0f, -1.0f},
    { 1.0f,  1.0f, -1.0f,   1.0f, 0.0f,    0.0f,  0.0f, -1.0f},
    { 1.0f, -1.0f, -1.0f,   1.0f, 1.0f,    0.0f,  0.0f, -1.0f},
    { 1.0f, -1.0f, -1.0f,   1.0f, 1.0f,    0.0f,  0.0f, -1.0f},
    {-1.0f, -1.0f, -1.0f,   0.0f, 1.0f,    0.0f,  0.0f, -1.0f},
    {-1.0f,  1.0f, -1.0f,   0.0f, 0.0f,    0.0f,  0.0f, -1.0f},

    // +Z face
    { 1.0f,  1.0f,  1.0f,   0.0f, 0.0f,    0.0f,  0.0f,  1.0f},
    {-1.0f,  1.0f,  1.0f,   1.0f, 0.0f,    0.0f,  0.0f,  1.0f},
    {-1.0f, -1.0f,  1.0f,   1.0f, 1.0f,    0.0f,  0.0f,  1.0f},
    {-1.0f, -1.0f,  1.0f,   1.0f, 1.0f,    0.0f,  0.0f,  1.0f},
    { 1.0f, -1.0f,  1.0f,   0.0f, 1.0f,    0.0f,  0.0f,  1.0f},
    { 1.0f,  1.0f,  1.0f,   0.0f, 0.0f,    0.0f,  0.0f,  1.0f},

    // -Y face
    {-1.0f, -1.0f, -1.0f,   0.0f, 0.0f,    0.0f, -1.0f,  0.0f},
    { 1.0f, -1.0f, -1.0f,   1.0f, 0.0f,    0.0f, -1.0f,  0.0f},
    { 1.0f, -1.0f,  1.0f,   1.0f, 1.0f,    0.0f, -1.0f,  0.0f},
    { 1.0f, -1.0f,  1.0f,   1.0f, 1.0f,    0.0f, -1.0f,  0.0f},
    {-1.0f, -1.0f,  1.0f,   0.0f, 1.0f,    0.0f, -1.0f,  0.0f},
    {-1.0f, -1.0f, -1.0f,   0.0f, 0.0f,    0.0f, -1.0f,  0.0f},

    // +Y face
    {-1.0f,  1.0f,  1.0f,   0.0f, 0.0f,    0.0f,  1.0f,  0.0f},
    { 1.0f,  1.0f,  1.0f,   1.0f, 0.0f,    0.0f,  1.0f,  0.0f},
    { 1.0f,  1.0f, -1.0f,   1.0f, 1.0f,    0.0f,  1.0f,  0.0f},
    { 1.0f,  1.0f, -1.0f,   1.0f, 1.0f,    0.0f,  1.0f,  0.0f},
    {-1.0f,  1.0f, -1.0f,   0.0f, 1.0f,    0.0f,  1.0f,  0.0f},
    {-1.0f,  1.0f,  1.0f,   0.0f, 0.0f,    0.0f,  1.0f,  0.0f},

    // -X face
    {-1.0f,  1.0f,  1.0f,   0.0f, 0.0f,   -1.0f,  0.0f,  0.0f},
    {-1.0f,  1.0f, -1.0f,   1.0f, 0.0f,   -1.0f,  0.0f,  0.0f},
    {-1.0f, -1.0f, -1.0f,   1.0f, 1.0f,   -1.0f,  0.0f,  0.0f},
    {-1.0f, -1.0f, -1.0f,   1.0f, 1.0f,   -1.0f,  0.0f,  0.0f},
    {-1.0f, -1.0f,  1.0f,   0.0f, 1.0f,   -1.0f,  0.0f,  0.0f},
    {-1.0f,  1.0f,  1.0f,   0.0f, 0.0f,   -1.0f,  0.0f,  0.0f},

    // +X face
    { 1.0f,  1.0f, -1.0f,   0.0f, 0.0f,    1.0f,  0.0f,  0.0f},
    { 1.0f,  1.0f,  1.0f,   1.0f, 0.0f,    1.0f,  0.0f,  0.0f},
    { 1.0f, -1.0f,  1.0f,   1.0f, 1.0f,    1.0f,  0.0f,  0.0f},
    { 1.0f, -1.0f,  1.0f,   1.0f, 1.0f,    1.0f,  0.0f,  0.0f},
    { 1.0f, -1.0f, -1.0f,   0.0f, 1.0f,    1.0f,  0.0f,  0.0f},
    { 1.0f,  1.0f, -1.0f,   0.0f, 0.0f,    1.0f,  0.0f,  0.0f} 
};

//-----------------------------------------------------------------------------
// Function Prototypes.
//-----------------------------------------------------------------------------

void    ChooseBestMSAAMode(D3DFORMAT backBufferFmt, D3DFORMAT depthStencilFmt,
                           BOOL windowed, D3DMULTISAMPLE_TYPE &type,
                           DWORD &qualityLevels, DWORD &samplesPerPixel);
void    Cleanup();
void    CleanupApp();
HWND    CreateAppWindow(const WNDCLASSEX &wcl, const char *pszTitle);
bool    CreateNullTexture(int width, int height, LPDIRECT3DTEXTURE9 &pTexture);
bool    DeviceIsValid();
float   GetElapsedTimeInSeconds();
LPCSTR  GetLightTypeStr();
bool    Init();
void    InitApp();
void    InitCube();
bool    InitD3D();
bool    InitFont(const char *pszFont, int ptSize, LPD3DXFONT &pFont);
bool    LoadShader(const char *pszFilename, LPD3DXEFFECT &pEffect);
void    Log(const char *pszMessage);
bool    MSAAModeSupported(D3DMULTISAMPLE_TYPE type, D3DFORMAT backBufferFmt,
                          D3DFORMAT depthStencilFmt, BOOL windowed,
                          DWORD &qualityLevels);
void    ProcessMouseInput(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
void    RenderCubeFixedFunc();
void    RenderCubeProgrammable();
void    RenderFrame();
void    RenderText();
bool    ResetDevice();
void    SetProcessorAffinity();
void    ToggleFullScreen();
void    UpdateEffect();
void    UpdateFrame(float elapsedTimeSec);
void    UpdateFrameRate(float elapsedTimeSec);
void    UpdateLighting();
LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

//-----------------------------------------------------------------------------
// Functions.
//-----------------------------------------------------------------------------

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
#if defined(_DEBUG)
    _CrtSetDbgFlag(_CRTDBG_LEAK_CHECK_DF | _CRTDBG_ALLOC_MEM_DF);
    _CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_FILE);
    _CrtSetReportFile(_CRT_ASSERT, _CRTDBG_FILE_STDERR);
#endif

    MSG msg = {0};
    WNDCLASSEX wcl = {0};

    wcl.cbSize = sizeof(wcl);
    wcl.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
    wcl.lpfnWndProc = WindowProc;
    wcl.cbClsExtra = 0;
    wcl.cbWndExtra = 0;
    wcl.hInstance = g_hInstance = hInstance;
    wcl.hIcon = LoadIcon(0, IDI_APPLICATION);
    wcl.hCursor = LoadCursor(0, IDC_ARROW);
    wcl.hbrBackground = 0;
    wcl.lpszMenuName = 0;
    wcl.lpszClassName = "D3D9WindowClass";
    wcl.hIconSm = 0;

    if (!RegisterClassEx(&wcl))
        return 0;

    g_hWnd = CreateAppWindow(wcl, APP_TITLE);

    if (g_hWnd)
    {
        SetProcessorAffinity();

        if (Init())
        {
            ShowWindow(g_hWnd, nShowCmd);
            UpdateWindow(g_hWnd);

            while (true)
            {
                while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
                {
                    if (msg.message == WM_QUIT)
                        break;

                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }

                if (msg.message == WM_QUIT)
                    break;

                if (g_hasFocus)
                {
                    UpdateFrame(GetElapsedTimeInSeconds());

                    if (DeviceIsValid())
                        RenderFrame();
                }
                else
                {
                    WaitMessage();
                }
            }
        }

        Cleanup();
        UnregisterClass(wcl.lpszClassName, hInstance);
    }

    return static_cast<int>(msg.wParam);
}

LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg)
    {
    case WM_ACTIVATE:
        switch (wParam)
        {
        default:
            break;

        case WA_ACTIVE:
        case WA_CLICKACTIVE:
            g_hasFocus = true;
            break;

        case WA_INACTIVE:
            if (g_isFullScreen)
                ShowWindow(hWnd, SW_MINIMIZE);
            g_hasFocus = false;
            break;
        }
        break;

    case WM_CHAR:
        switch (static_cast<int>(wParam))
        {
        case VK_ESCAPE:
            PostMessage(hWnd, WM_CLOSE, 0, 0);
            break;
        
        case VK_RETURN:
            g_fixedFunction = !g_fixedFunction;
            break;

        case VK_SPACE:
            if (++g_lightType > SPOT_LIGHT)
                g_lightType = DIR_LIGHT;

            UpdateLighting();
            break;

        case VK_BACK:
            if (--g_lightType < DIR_LIGHT)
                g_lightType = SPOT_LIGHT;

            UpdateLighting();
            break;

        case 'h':
        case 'H':
            g_displayHelp = !g_displayHelp;
            break;

        case 't':
        case 'T':
            g_disableColorMapTexture = !g_disableColorMapTexture;
            break;

        default:
            break;
        }
        break;

    case WM_DESTROY:
        PostQuitMessage(0);
        return 0;

    case WM_SIZE:
        g_windowWidth = static_cast<int>(LOWORD(lParam));
        g_windowHeight = static_cast<int>(HIWORD(lParam));
        break;

    case WM_SYSKEYDOWN:
        if (wParam == VK_RETURN)
            ToggleFullScreen();
        break;

    default:
        ProcessMouseInput(hWnd, msg, wParam, lParam);
        break;
    }

    return DefWindowProc(hWnd, msg, wParam, lParam);
}

void ChooseBestMSAAMode(D3DFORMAT backBufferFmt, D3DFORMAT depthStencilFmt,
                        BOOL windowed, D3DMULTISAMPLE_TYPE &type,
                        DWORD &qualityLevels, DWORD &samplesPerPixel)
{
    bool supported = false;

    struct MSAAMode
    {
        D3DMULTISAMPLE_TYPE type;
        DWORD samples;
    }
    multsamplingTypes[15] =
    {
        { D3DMULTISAMPLE_16_SAMPLES,  16 },
        { D3DMULTISAMPLE_15_SAMPLES,  15 },
        { D3DMULTISAMPLE_14_SAMPLES,  14 },
        { D3DMULTISAMPLE_13_SAMPLES,  13 },
        { D3DMULTISAMPLE_12_SAMPLES,  12 },
        { D3DMULTISAMPLE_11_SAMPLES,  11 },
        { D3DMULTISAMPLE_10_SAMPLES,  10 },
        { D3DMULTISAMPLE_9_SAMPLES,   9  },
        { D3DMULTISAMPLE_8_SAMPLES,   8  },
        { D3DMULTISAMPLE_7_SAMPLES,   7  },
        { D3DMULTISAMPLE_6_SAMPLES,   6  },
        { D3DMULTISAMPLE_5_SAMPLES,   5  },
        { D3DMULTISAMPLE_4_SAMPLES,   4  },
        { D3DMULTISAMPLE_3_SAMPLES,   3  },
        { D3DMULTISAMPLE_2_SAMPLES,   2  }
    };

    for (int i = 0; i < 15; ++i)
    {
        type = multsamplingTypes[i].type;

        supported = MSAAModeSupported(type, backBufferFmt, depthStencilFmt,
                        windowed, qualityLevels);

        if (supported)
        {
            samplesPerPixel = multsamplingTypes[i].samples;
            return;
        }
    }

    type = D3DMULTISAMPLE_NONE;
    qualityLevels = 0;
    samplesPerPixel = 1;
}

void Cleanup()
{
    CleanupApp();

    SAFE_RELEASE(g_pNullTexture);
    SAFE_RELEASE(g_pFont);
    SAFE_RELEASE(g_pDevice);
    SAFE_RELEASE(g_pDirect3D);
}

void CleanupApp()
{
    SAFE_RELEASE(g_pEffect);
    SAFE_RELEASE(g_pColorMapTexture);
    SAFE_RELEASE(g_pVertexBuffer);
    SAFE_RELEASE(g_pVertexDecl);
}

HWND CreateAppWindow(const WNDCLASSEX &wcl, const char *pszTitle)
{
    // Create a window that is centered on the desktop. It's exactly 1/4 the
    // size of the desktop. Don't allow it to be resized.

    DWORD wndExStyle = WS_EX_OVERLAPPEDWINDOW;
    DWORD wndStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
                     WS_MINIMIZEBOX | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;

    HWND hWnd = CreateWindowEx(wndExStyle, wcl.lpszClassName, pszTitle,
                    wndStyle, 0, 0, 0, 0, 0, 0, wcl.hInstance, 0);

    if (hWnd)
    {
        int screenWidth = GetSystemMetrics(SM_CXSCREEN);
        int screenHeight = GetSystemMetrics(SM_CYSCREEN);
        int halfScreenWidth = screenWidth / 2;
        int halfScreenHeight = screenHeight / 2;
        int left = (screenWidth - halfScreenWidth) / 2;
        int top = (screenHeight - halfScreenHeight) / 2;
        RECT rc = {0};

        SetRect(&rc, left, top, left + halfScreenWidth, top + halfScreenHeight);
        AdjustWindowRectEx(&rc, wndStyle, FALSE, wndExStyle);
        MoveWindow(hWnd, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, TRUE);

        GetClientRect(hWnd, &rc);
        g_windowWidth = rc.right - rc.left;
        g_windowHeight = rc.bottom - rc.top;
    }

    return hWnd;
}

bool CreateNullTexture(int width, int height, LPDIRECT3DTEXTURE9 &pTexture)
{
    // Create an empty white texture. This texture is applied to geometry
    // that doesn't have any texture maps. This trick allows the same shader to
    // be used to draw the geometry with and without textures applied.

    HRESULT hr = D3DXCreateTexture(g_pDevice, width, height, 0, 0,
                    D3DFMT_X8R8G8B8, D3DPOOL_MANAGED, &pTexture);

    if (FAILED(hr))
        return false;

    LPDIRECT3DSURFACE9 pSurface = 0;

    if (SUCCEEDED(pTexture->GetSurfaceLevel(0, &pSurface)))
    {
        D3DLOCKED_RECT rcLock = {0};

        if (SUCCEEDED(pSurface->LockRect(&rcLock, 0, 0)))
        {
            BYTE *pPixels = static_cast<BYTE*>(rcLock.pBits);
            int widthInBytes = width * 4;

            if (widthInBytes == rcLock.Pitch)
            {
                memset(pPixels, 0xff, widthInBytes * height);
            }
            else
            {
                for (int y = 0; y < height; ++y)
                    memset(&pPixels[y * rcLock.Pitch], 0xff, rcLock.Pitch);
            }

            pSurface->UnlockRect();
            pSurface->Release();
            return true;
        }

        pSurface->Release();
    }

    pTexture->Release();
    return false;
}

bool DeviceIsValid()
{
    HRESULT hr = g_pDevice->TestCooperativeLevel();

    if (FAILED(hr))
    {
        if (hr == D3DERR_DEVICENOTRESET)
            return ResetDevice();
    }

    return true;
}

float GetElapsedTimeInSeconds()
{
    // Returns the elapsed time (in seconds) since the last time this function
    // was called. This elaborate setup is to guard against large spikes in
    // the time returned by QueryPerformanceCounter().

    static const int MAX_SAMPLE_COUNT = 50;

    static float frameTimes[MAX_SAMPLE_COUNT];
    static float timeScale = 0.0f;
    static float actualElapsedTimeSec = 0.0f;
    static INT64 freq = 0;
    static INT64 lastTime = 0;
    static int sampleCount = 0;
    static bool initialized = false;

    INT64 time = 0;
    float elapsedTimeSec = 0.0f;

    if (!initialized)
    {
        initialized = true;
        QueryPerformanceFrequency(reinterpret_cast<LARGE_INTEGER*>(&freq));
        QueryPerformanceCounter(reinterpret_cast<LARGE_INTEGER*>(&lastTime));
        timeScale = 1.0f / freq;
    }

    QueryPerformanceCounter(reinterpret_cast<LARGE_INTEGER*>(&time));
    elapsedTimeSec = (time - lastTime) * timeScale;
    lastTime = time;

    if (fabsf(elapsedTimeSec - actualElapsedTimeSec) < 1.0f)
    {
        memmove(&frameTimes[1], frameTimes, sizeof(frameTimes) - sizeof(frameTimes[0]));
        frameTimes[0] = elapsedTimeSec;

        if (sampleCount < MAX_SAMPLE_COUNT)
            ++sampleCount;
    }

    actualElapsedTimeSec = 0.0f;

    for (int i = 0; i < sampleCount; ++i)
        actualElapsedTimeSec += frameTimes[i];

    if (sampleCount > 0)
        actualElapsedTimeSec /= sampleCount;

    return actualElapsedTimeSec;
}

LPCSTR GetLightTypeStr()
{
    switch (g_lightType)
    {
    case DIR_LIGHT:
        return "Directional lighting";

    case POINT_LIGHT:
        return "Point lighting";

    case SPOT_LIGHT:
        return "Spot lighting";

    default:
        return "Unknown lighting";
    }
}

bool Init()
{
    if (!InitD3D())
    {
        Log("Direct3D initialization failed!");
        return false;
    }

    try
    {
        InitApp();
        return true;
    }
    catch (const std::exception &e)
    {
        std::ostringstream msg;

        msg << "Application initialization failed!" << std::endl << std::endl;
        msg << e.what();

        Log(msg.str().c_str());
        return false;
    }
}

void InitApp()
{
    if (!InitFont("Arial", 10, g_pFont))
        throw std::runtime_error("Failed to create font.");

    if (!CreateNullTexture(2, 2, g_pNullTexture))
        throw std::runtime_error("Failed to create null texture.");

    if (FAILED(D3DXCreateTextureFromFile(g_pDevice, "color_map.dds", &g_pColorMapTexture)))
        throw std::runtime_error("Failed to load texture: color_map.dds.");

    if (!LoadShader("vertex_lighting.fx", g_pEffect))
        throw std::runtime_error("Failed to load shader: vertex_lighting.fx.");

    InitCube();
}

void InitCube()
{
    // Create the vertex declaration for the cube.

    if (FAILED(g_pDevice->CreateVertexDeclaration(g_vertexElements, &g_pVertexDecl)))
        throw std::runtime_error("Failed to create vertex declaration.");

    // Create a vertex buffer for the cube.

    if (FAILED(g_pDevice->CreateVertexBuffer(sizeof(Vertex) * 36, 0, 0,
        D3DPOOL_MANAGED, &g_pVertexBuffer, 0)))
        throw std::runtime_error("Failed to create vertex buffer.");

    // Fill vertex buffer with the cube's geometry.

    Vertex *pVertices = 0;

    if (FAILED(g_pVertexBuffer->Lock(0, 0, reinterpret_cast<void**>(&pVertices), 0)))
        throw std::runtime_error("Failed to lock vertex buffer.");

    memcpy(pVertices, g_cube, sizeof(g_cube));
    g_pVertexBuffer->Unlock();
}

bool InitD3D()
{
    HRESULT hr = 0;
    D3DDISPLAYMODE desktop = {0};

    g_pDirect3D = Direct3DCreate9(D3D_SDK_VERSION);

    if (!g_pDirect3D)
        return false;

    // Just use the current desktop display mode.
    hr = g_pDirect3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &desktop);

    if (FAILED(hr))
    {
        g_pDirect3D->Release();
        g_pDirect3D = 0;
        return false;
    }

    // Setup Direct3D for windowed rendering.
    g_params.BackBufferWidth = 0;
    g_params.BackBufferHeight = 0;
    g_params.BackBufferFormat = desktop.Format;
    g_params.BackBufferCount = 1;
    g_params.hDeviceWindow = g_hWnd;
    g_params.Windowed = TRUE;
    g_params.EnableAutoDepthStencil = TRUE;
    g_params.AutoDepthStencilFormat = D3DFMT_D24S8;
    g_params.Flags = D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
    g_params.FullScreen_RefreshRateInHz = 0;

    if (g_enableVerticalSync)
        g_params.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
    else
        g_params.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;

    // Swap effect must be D3DSWAPEFFECT_DISCARD for multi-sampling support.
    g_params.SwapEffect = D3DSWAPEFFECT_DISCARD;

    // Select the highest quality multi-sample anti-aliasing (MSAA) mode.
    ChooseBestMSAAMode(g_params.BackBufferFormat, g_params.AutoDepthStencilFormat,
        g_params.Windowed, g_params.MultiSampleType, g_params.MultiSampleQuality,
        g_msaaSamples);

    // Most modern video cards should have no problems creating pure devices.
    // Note that by creating a pure device we lose the ability to debug vertex
    // and pixel shaders.
    hr = g_pDirect3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_hWnd,
            D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE,
            &g_params, &g_pDevice);

    if (FAILED(hr))
    {
        // Fall back to software vertex processing for less capable hardware.
        // Note that in order to debug vertex shaders we must use a software
        // vertex processing device.
        hr = g_pDirect3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_hWnd,
                D3DCREATE_SOFTWARE_VERTEXPROCESSING, &g_params, &g_pDevice);
    }

    if (FAILED(hr))
    {
        g_pDirect3D->Release();
        g_pDirect3D = 0;
        return false;
    }

    D3DCAPS9 caps;

    // Prefer anisotropic texture filtering if it's supported.
    if (SUCCEEDED(g_pDevice->GetDeviceCaps(&caps)))
    {
        if (caps.RasterCaps & D3DPRASTERCAPS_ANISOTROPY)
            g_maxAnisotrophy = caps.MaxAnisotropy;
        else
            g_maxAnisotrophy = 1;
    }

    return true;
}

bool InitFont(const char *pszFont, int ptSize, LPD3DXFONT &pFont)
{
    static DWORD dwQuality = 0;

    // Prefer ClearType font quality if available.

    if (!dwQuality)
    {
        DWORD dwVersion = GetVersion();
        DWORD dwMajorVersion = static_cast<DWORD>((LOBYTE(LOWORD(dwVersion))));
        DWORD dwMinorVersion = static_cast<DWORD>((HIBYTE(LOWORD(dwVersion))));

        // Windows XP and higher will support ClearType quality fonts.
        if (dwMajorVersion >= 6 || (dwMajorVersion == 5 && dwMinorVersion == 1))
            dwQuality = CLEARTYPE_QUALITY;
        else
            dwQuality = ANTIALIASED_QUALITY;
    }

    int logPixelsY = 0;

    // Convert from font point size to pixel size.

    if (HDC hDC = GetDC((0)))
    {
        logPixelsY = GetDeviceCaps(hDC, LOGPIXELSY);
        ReleaseDC(0, hDC);
    }

    int fontCharHeight = -ptSize * logPixelsY / 72;

    // Now create the font. Prefer anti-aliased text.

    HRESULT hr = D3DXCreateFont(
        g_pDevice,
        fontCharHeight,                 // height
        0,                              // width
        FW_BOLD,                        // weight
        1,                              // mipmap levels
        FALSE,                          // italic
        DEFAULT_CHARSET,                // char set
        OUT_DEFAULT_PRECIS,             // output precision
        dwQuality,                      // quality
        DEFAULT_PITCH | FF_DONTCARE,    // pitch and family
        pszFont,                        // font name
        &pFont);

    return SUCCEEDED(hr) ? true : false;
}

bool LoadShader(const char *pszFilename, LPD3DXEFFECT &pEffect)
{
    ID3DXBuffer *pCompilationErrors = 0;
    DWORD dwShaderFlags = D3DXFX_NOT_CLONEABLE | D3DXSHADER_NO_PRESHADER;

    // Both vertex and pixel shaders can be debugged. To enable shader
    // debugging add the following flag to the dwShaderFlags variable:
    //      dwShaderFlags |= D3DXSHADER_DEBUG;
    //
    // Vertex shaders can be debugged with either the REF device or a device
    // created for software vertex processing (i.e., the IDirect3DDevice9
    // object must be created with the D3DCREATE_SOFTWARE_VERTEXPROCESSING
    // behavior). Pixel shaders can be debugged only using the REF device.
    //
    // To enable vertex shader debugging add the following flag to the
    // dwShaderFlags variable:
    //     dwShaderFlags |= D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
    //
    // To enable pixel shader debugging add the following flag to the
    // dwShaderFlags variable:
    //     dwShaderFlags |= D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;

    HRESULT hr = D3DXCreateEffectFromFile(g_pDevice, pszFilename, 0, 0,
                    dwShaderFlags, 0, &pEffect, &pCompilationErrors);

    if (FAILED(hr))
    {
        if (pCompilationErrors)
        {
            std::string compilationErrors(static_cast<const char *>(
                            pCompilationErrors->GetBufferPointer()));

            pCompilationErrors->Release();
            throw std::runtime_error(compilationErrors);
        }
    }

    if (pCompilationErrors)
        pCompilationErrors->Release();

    return pEffect != 0;
}

void Log(const char *pszMessage)
{
    MessageBox(0, pszMessage, "Error", MB_ICONSTOP);
}

bool MSAAModeSupported(D3DMULTISAMPLE_TYPE type, D3DFORMAT backBufferFmt,
                       D3DFORMAT depthStencilFmt, BOOL windowed,
                       DWORD &qualityLevels)
{
    DWORD backBufferQualityLevels = 0;
    DWORD depthStencilQualityLevels = 0;

    HRESULT hr = g_pDirect3D->CheckDeviceMultiSampleType(D3DADAPTER_DEFAULT,
                    D3DDEVTYPE_HAL, backBufferFmt, windowed, type,
                    &backBufferQualityLevels);

    if (SUCCEEDED(hr))
    {
        hr = g_pDirect3D->CheckDeviceMultiSampleType(D3DADAPTER_DEFAULT,
                D3DDEVTYPE_HAL, depthStencilFmt, windowed, type,
                &depthStencilQualityLevels);

        if (SUCCEEDED(hr))
        {
            if (backBufferQualityLevels == depthStencilQualityLevels)
            {
                // The valid range is between zero and one less than the level
                // returned by IDirect3D9::CheckDeviceMultiSampleType().

                if (backBufferQualityLevels > 0)
                    qualityLevels = backBufferQualityLevels - 1;
                else
                    qualityLevels = backBufferQualityLevels;

                return true;
            }
        }
    }

    return false;
}

void ProcessMouseInput(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    // Use the left mouse button to track the camera.
    // Use the middle mouse button to dolly the camera.
    // Use the right mouse button to orbit the camera.

    enum CameraMode {CAMERA_NONE, CAMERA_TRACK, CAMERA_DOLLY, CAMERA_ORBIT};

    static CameraMode cameraMode = CAMERA_NONE;
    static POINT ptMousePrev = {0};
    static POINT ptMouseCurrent = {0};
    static int mouseButtonsDown = 0;
    static float dx = 0.0f;
    static float dy = 0.0f;
    static float wheelDelta = 0.0f;

    switch (msg)
    {
    case WM_LBUTTONDOWN:
        cameraMode = CAMERA_TRACK;
        ++mouseButtonsDown;
        SetCapture(hWnd);
        ptMousePrev.x = static_cast<int>(static_cast<short>(LOWORD(lParam)));
        ptMousePrev.y = static_cast<int>(static_cast<short>(HIWORD(lParam)));
        ClientToScreen(hWnd, &ptMousePrev);
        break;

    case WM_RBUTTONDOWN:
        cameraMode = CAMERA_ORBIT;
        ++mouseButtonsDown;
        SetCapture(hWnd);
        ptMousePrev.x = static_cast<int>(static_cast<short>(LOWORD(lParam)));
        ptMousePrev.y = static_cast<int>(static_cast<short>(HIWORD(lParam)));
        ClientToScreen(hWnd, &ptMousePrev);
        break;

    case WM_MBUTTONDOWN:
        cameraMode = CAMERA_DOLLY;
        ++mouseButtonsDown;
        SetCapture(hWnd);
        ptMousePrev.x = static_cast<int>(static_cast<short>(LOWORD(lParam)));
        ptMousePrev.y = static_cast<int>(static_cast<short>(HIWORD(lParam)));
        ClientToScreen(hWnd, &ptMousePrev);
        break;

    case WM_MOUSEMOVE:
        ptMouseCurrent.x = static_cast<int>(static_cast<short>(LOWORD(lParam)));
        ptMouseCurrent.y = static_cast<int>(static_cast<short>(HIWORD(lParam)));
        ClientToScreen(hWnd, &ptMouseCurrent);

        switch (cameraMode)
        {
        case CAMERA_TRACK:
            dx = static_cast<float>(ptMouseCurrent.x - ptMousePrev.x);
            dx *= MOUSE_TRACK_SPEED;

            dy = static_cast<float>(ptMouseCurrent.y - ptMousePrev.y);
            dy *= MOUSE_TRACK_SPEED;

            g_cameraPos.x -= dx;
            g_cameraPos.y += dy;

            g_targetPos.x -= dx;
            g_targetPos.y += dy;

            UpdateLighting();
            break;

        case CAMERA_DOLLY:
            dy = static_cast<float>(ptMousePrev.y - ptMouseCurrent.y);
            dy *= MOUSE_DOLLY_SPEED;

            g_cameraPos.z += dy;

            if (g_cameraPos.z > DOLLY_MAX)
                g_cameraPos.z = DOLLY_MAX;

            if (g_cameraPos.z < DOLLY_MIN)
                g_cameraPos.z = DOLLY_MIN;

            UpdateLighting();
            break;

        case CAMERA_ORBIT:
            dx = static_cast<float>(ptMousePrev.x - ptMouseCurrent.x);
            dx *= MOUSE_ORBIT_SPEED;

            dy = static_cast<float>(ptMousePrev.y - ptMouseCurrent.y);
            dy *= MOUSE_ORBIT_SPEED;

            g_heading += dx;
            g_pitch += dy;

            if (g_pitch > 90.0f)
                g_pitch = 90.0f;

            if (g_pitch < -90.0f)
                g_pitch = -90.0f;

            break;
        }

        ptMousePrev.x = ptMouseCurrent.x;
        ptMousePrev.y = ptMouseCurrent.y;
        break;

    case WM_LBUTTONUP:
    case WM_RBUTTONUP:
    case WM_MBUTTONUP:
        if (--mouseButtonsDown <= 0)
        {
            mouseButtonsDown = 0;
            cameraMode = CAMERA_NONE;
            ReleaseCapture();
        }
        else
        {
            if (wParam & MK_LBUTTON)
                cameraMode = CAMERA_TRACK;
            else if (wParam & MK_RBUTTON)
                cameraMode = CAMERA_ORBIT;
            else if (wParam & MK_MBUTTON)
                cameraMode = CAMERA_DOLLY;
        }
        break;

    case WM_MOUSEWHEEL:
        wheelDelta = static_cast<float>(static_cast<int>(wParam) >> 16);
        g_cameraPos.z += wheelDelta * MOUSE_WHEEL_DOLLY_SPEED;

        if (g_cameraPos.z > DOLLY_MAX)
            g_cameraPos.z = DOLLY_MAX;

        if (g_cameraPos.z < DOLLY_MIN)
            g_cameraPos.z = DOLLY_MIN;

        UpdateLighting();
        break;

    default:
        break;
    }
}

void RenderCubeFixedFunc()
{
    static D3DXMATRIX xRot, yRot;
    static D3DXMATRIX world, view, proj;

    // Setup textures. Prefer anisotropic filtering if available.

    if (g_disableColorMapTexture)
        g_pDevice->SetTexture(0, g_pNullTexture);
    else
        g_pDevice->SetTexture(0, g_pColorMapTexture);

    g_pDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
    g_pDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);

    if (g_maxAnisotrophy > 1)
    {
        g_pDevice->SetSamplerState(0, D3DSAMP_MAXANISOTROPY, g_maxAnisotrophy);
        g_pDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_ANISOTROPIC);
    }
    else
    {
        g_pDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
    }

    // Define global ambient term for the scene.

    g_pDevice->SetRenderState(D3DRS_AMBIENT, D3DCOLOR_COLORVALUE(
        g_sceneAmbient[0], g_sceneAmbient[1], g_sceneAmbient[2], 1.0f));

    // Calculate matrices. Projection, view, and world matrices.

    D3DXMatrixPerspectiveFovLH(&proj, CAMERA_FOVY,
        static_cast<float>(g_windowWidth) / static_cast<float>(g_windowHeight),
        CAMERA_ZNEAR, CAMERA_ZFAR);

    D3DXMatrixLookAtLH(&view, &g_cameraPos, &g_targetPos, &D3DXVECTOR3(0.0f, 1.0f, 0.0f));

    D3DXMatrixRotationX(&xRot, D3DXToRadian(g_pitch));
    D3DXMatrixRotationY(&yRot, D3DXToRadian(g_heading));
    D3DXMatrixMultiply(&world, &yRot, &xRot);

    // Setup matrices.

    g_pDevice->SetTransform(D3DTS_PROJECTION, &proj);
    g_pDevice->SetTransform(D3DTS_VIEW, &view);
    g_pDevice->SetTransform(D3DTS_WORLD, &world);

    // Setup lighting and material.

    g_pDevice->SetMaterial(&g_material);
    g_pDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
    g_pDevice->SetLight(0, &g_light);
    g_pDevice->LightEnable(0, TRUE);

    // Render the scene.

    g_pDevice->SetVertexDeclaration(g_pVertexDecl);
    g_pDevice->SetStreamSource(0, g_pVertexBuffer, 0, sizeof(Vertex));

    g_pDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 12);

    g_pDevice->SetStreamSource(0, 0, 0, 0);
    g_pDevice->SetTexture(0, 0);
}

void RenderCubeProgrammable()
{
    static UINT totalPasses;
    static D3DXHANDLE hTechnique;

    // Select the shader to use based on current lighting type.
    switch (g_lightType)
    {
    case DIR_LIGHT:
        hTechnique = g_pEffect->GetTechniqueByName("PerVertexDirectionalLighting");
        break;

    case POINT_LIGHT:
        hTechnique = g_pEffect->GetTechniqueByName("PerVertexPointLighting");
        break;

    case SPOT_LIGHT:
        hTechnique = g_pEffect->GetTechniqueByName("PerVertexSpotLighting");
        break;

    default:
        hTechnique = 0;
        break;
    }

    // Bind vertex buffer.
    g_pDevice->SetVertexDeclaration(g_pVertexDecl);
    g_pDevice->SetStreamSource(0, g_pVertexBuffer, 0, sizeof(Vertex));

    // Render the cube.
    if (SUCCEEDED(g_pEffect->SetTechnique(hTechnique)))
    {
        if (SUCCEEDED(g_pEffect->Begin(&totalPasses, 0)))
        {
            for (UINT pass = 0; pass < totalPasses; ++pass)
            {
                if (SUCCEEDED(g_pEffect->BeginPass(pass)))
                {
                    g_pDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 12);
                    g_pEffect->EndPass();
                }
            }

            g_pEffect->End();
        }
    }

    g_pDevice->SetStreamSource(0, 0, 0, 0);
    g_pDevice->SetTexture(0, 0);
}

void RenderFrame()
{
    g_pDevice->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
        D3DCOLOR_XRGB(100, 149, 237),   // CornflowerBlue
        1.0f, 0);

    if (FAILED(g_pDevice->BeginScene()))
        return;

    if (g_fixedFunction)
        RenderCubeFixedFunc();
    else
        RenderCubeProgrammable();

    RenderText();

    g_pDevice->EndScene();
    g_pDevice->Present(0, 0, 0, 0);
}

void RenderText()
{
    static RECT rcClient;

    std::ostringstream output;

    if (g_displayHelp)
    {
        output
            << "Left mouse click and drag to track camera" << std::endl
            << "Middle mouse click and drag to dolly camera" << std::endl
            << "Right mouse click and drag to orbit camera" << std::endl
            << "Mouse wheel to dolly camera" << std::endl
            << std::endl
            << "Press ENTER to toggle between fixed function and programmable pipelines" << std::endl
            << "Press SPACE to cycle forwards through light types" << std::endl
            << "Press BACKSPACE to cycle backwards through light types" << std::endl
            << "Press T to enable/disable textures" << std::endl
            << "Press ALT + ENTER to toggle full screen" << std::endl
            << "Press ESC to exit" << std::endl
            << std::endl
            << "Press H to hide help";
    }
    else
    {
        output
            << "FPS: " << g_framesPerSecond << std::endl
            << "Multisample anti-aliasing: " << g_msaaSamples << "x" << std::endl
            << "Anisotropic filtering: " << g_maxAnisotrophy << "x" << std::endl
            << GetLightTypeStr() << std::endl
            << (g_fixedFunction ? "Fixed function pipeline" : "Programmable pipeline") << std::endl
            << std::endl
            << "Press H to display help";
    }

    GetClientRect(g_hWnd, &rcClient);
    rcClient.left += 4;
    rcClient.top += 2;

    g_pFont->DrawText(0, output.str().c_str(), -1, &rcClient,
        DT_EXPANDTABS | DT_LEFT, D3DCOLOR_XRGB(255, 255, 0));
}

bool ResetDevice()
{
    if (FAILED(g_pEffect->OnLostDevice()))
        return false;

    if (FAILED(g_pFont->OnLostDevice()))
        return false;

    if (FAILED(g_pDevice->Reset(&g_params)))
        return false;

    if (FAILED(g_pFont->OnResetDevice()))
        return false;

    if (FAILED(g_pEffect->OnResetDevice()))
        return false;

    return true;
}

void SetProcessorAffinity()
{
    // Assign the current thread to one processor. This ensures that timing
    // code runs on only one processor, and will not suffer any ill effects
    // from power management.
    //
    // Based on the DXUTSetProcessorAffinity() function in the DXUT framework.

    DWORD_PTR dwProcessAffinityMask = 0;
    DWORD_PTR dwSystemAffinityMask = 0;
    HANDLE hCurrentProcess = GetCurrentProcess();

    if (!GetProcessAffinityMask(hCurrentProcess, &dwProcessAffinityMask, &dwSystemAffinityMask))
        return;

    if (dwProcessAffinityMask)
    {
        // Find the lowest processor that our process is allowed to run against.

        DWORD_PTR dwAffinityMask = (dwProcessAffinityMask & ((~dwProcessAffinityMask) + 1));

        // Set this as the processor that our thread must always run against.
        // This must be a subset of the process affinity mask.

        HANDLE hCurrentThread = GetCurrentThread();

        if (hCurrentThread != INVALID_HANDLE_VALUE)
        {
            SetThreadAffinityMask(hCurrentThread, dwAffinityMask);
            CloseHandle(hCurrentThread);
        }
    }

    CloseHandle(hCurrentProcess);
}

void ToggleFullScreen()
{
    static DWORD savedExStyle;
    static DWORD savedStyle;
    static RECT rcSaved;

    g_isFullScreen = !g_isFullScreen;

    if (g_isFullScreen)
    {
        // Moving to full screen mode.

        savedExStyle = GetWindowLong(g_hWnd, GWL_EXSTYLE);
        savedStyle = GetWindowLong(g_hWnd, GWL_STYLE);
        GetWindowRect(g_hWnd, &rcSaved);

        SetWindowLong(g_hWnd, GWL_EXSTYLE, 0);
        SetWindowLong(g_hWnd, GWL_STYLE, WS_POPUP);
        SetWindowPos(g_hWnd, HWND_TOPMOST, 0, 0, 0, 0,
            SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED | SWP_SHOWWINDOW);

        g_windowWidth = GetSystemMetrics(SM_CXSCREEN);
        g_windowHeight = GetSystemMetrics(SM_CYSCREEN);

        SetWindowPos(g_hWnd, HWND_TOPMOST, 0, 0,
            g_windowWidth, g_windowHeight, SWP_SHOWWINDOW);

        // Update presentation parameters.

        g_params.Windowed = FALSE;
        g_params.BackBufferWidth = g_windowWidth;
        g_params.BackBufferHeight = g_windowHeight;

        if (g_enableVerticalSync)
        {
            g_params.FullScreen_RefreshRateInHz = D3DPRESENT_INTERVAL_DEFAULT;
            g_params.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
        }
        else
        {
            g_params.FullScreen_RefreshRateInHz = D3DPRESENT_INTERVAL_IMMEDIATE;
            g_params.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
        }
    }
    else
    {
        // Moving back to windowed mode.

        SetWindowLong(g_hWnd, GWL_EXSTYLE, savedExStyle);
        SetWindowLong(g_hWnd, GWL_STYLE, savedStyle);
        SetWindowPos(g_hWnd, HWND_NOTOPMOST, 0, 0, 0, 0,
            SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED | SWP_SHOWWINDOW);

        g_windowWidth = rcSaved.right - rcSaved.left;
        g_windowHeight = rcSaved.bottom - rcSaved.top;

        SetWindowPos(g_hWnd, HWND_NOTOPMOST, rcSaved.left, rcSaved.top,
            g_windowWidth, g_windowHeight, SWP_SHOWWINDOW);

        // Update presentation parameters.

        g_params.Windowed = TRUE;
        g_params.BackBufferWidth = g_windowWidth;
        g_params.BackBufferHeight = g_windowHeight;
        g_params.FullScreen_RefreshRateInHz = 0;

        if (g_enableVerticalSync)
            g_params.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
        else
            g_params.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
    }

    ResetDevice();
}

void UpdateEffect()
{
    static D3DXMATRIX xRot, yRot;
    static D3DXMATRIX world, view, proj;
    static D3DXMATRIX worldViewMatrix;
    static D3DXMATRIX worldViewProjectionMatrix;
    static D3DXMATRIX worldViewInverseTransposeMatrix;
    static float attenuation[3];

    D3DXMatrixPerspectiveFovLH(&proj, CAMERA_FOVY,
        static_cast<float>(g_windowWidth) / static_cast<float>(g_windowHeight),
        CAMERA_ZNEAR, CAMERA_ZFAR);

    D3DXMatrixLookAtLH(&view, &g_cameraPos, &g_targetPos, &D3DXVECTOR3(0.0f, 1.0f, 0.0f));

    // Calculate world matrix.
    D3DXMatrixRotationX(&xRot, D3DXToRadian(g_pitch));
    D3DXMatrixRotationY(&yRot, D3DXToRadian(g_heading));
    D3DXMatrixMultiply(&world, &yRot, &xRot);

    // Calculate combined world-view-projection matrix.
    worldViewProjectionMatrix = world * view * proj;

    // Calculate combined world-view matrix.
    worldViewMatrix = world * view;

    // Calculate inverse-transpose world-view matrix.
    D3DXMatrixInverse(&worldViewInverseTransposeMatrix, 0, &worldViewMatrix);
    D3DXMatrixTranspose(&worldViewInverseTransposeMatrix, &worldViewInverseTransposeMatrix);

    // Set the matrices for the shader.
    g_pEffect->SetMatrix("worldViewProjectionMatrix", &worldViewProjectionMatrix);
    g_pEffect->SetMatrix("worldViewMatrix", &worldViewMatrix);
    g_pEffect->SetMatrix("worldViewInverseTransposeMatrix", &worldViewInverseTransposeMatrix);
    g_pEffect->SetMatrix("viewMatrix", &view);

    // Set the scene global ambient term.
    g_pEffect->SetValue("globalAmbient", &g_sceneAmbient, sizeof(g_sceneAmbient));

    // Set the lighting parameters for the shader.
    attenuation[0] = g_light.Attenuation0;
    attenuation[1] = g_light.Attenuation1;
    attenuation[2] = g_light.Attenuation2;
    g_pEffect->SetValue("light.dir", &g_light.Direction, sizeof(g_light.Direction));
    g_pEffect->SetValue("light.pos", &g_light.Position, sizeof(g_light.Position));
    g_pEffect->SetValue("light.ambient", &g_light.Ambient, sizeof(g_light.Ambient));
    g_pEffect->SetValue("light.diffuse", &g_light.Diffuse, sizeof(g_light.Diffuse));
    g_pEffect->SetValue("light.specular", &g_light.Specular, sizeof(g_light.Specular));
    g_pEffect->SetValue("light.attenuation", attenuation, sizeof(attenuation));
    g_pEffect->SetFloat("light.falloff", g_light.Falloff);
    g_pEffect->SetFloat("light.theta", g_light.Theta);
    g_pEffect->SetFloat("light.phi", g_light.Phi);
    
    // Set the material parameters for the shader.
    g_pEffect->SetValue("material.ambient", &g_material.Ambient, sizeof(g_material.Ambient));
    g_pEffect->SetValue("material.diffuse", &g_material.Diffuse, sizeof(g_material.Diffuse));
    g_pEffect->SetValue("material.emissive", &g_material.Emissive, sizeof(g_material.Emissive));
    g_pEffect->SetValue("material.specular", &g_material.Specular, sizeof(g_material.Specular));
    g_pEffect->SetFloat("material.shininess", g_material.Power);

    // Bind the textures to the shader.
    if (g_disableColorMapTexture)
        g_pEffect->SetTexture("colorMapTexture", g_pNullTexture);
    else
        g_pEffect->SetTexture("colorMapTexture", g_pColorMapTexture);
}

void UpdateFrame(float elapsedTimeSec)
{
    UpdateFrameRate(elapsedTimeSec);

    if (!g_fixedFunction)
        UpdateEffect();
}

void UpdateFrameRate(float elapsedTimeSec)
{
    static float accumTimeSec = 0.0f;
    static int frames = 0;

    accumTimeSec += elapsedTimeSec;

    if (accumTimeSec > 1.0f)
    {
        g_framesPerSecond = frames;

        frames = 0;
        accumTimeSec = 0.0f;
    }
    else
    {
        ++frames;
    }
}

void UpdateLighting()
{
    switch (g_lightType)
    {
    case DIR_LIGHT:
        g_light.Type = D3DLIGHT_DIRECTIONAL;
        g_light.Direction.x = 0.0f;
        g_light.Direction.y = 0.0f;
        g_light.Direction.z = 1.0f;
        break;

    case POINT_LIGHT:
        g_light.Type = D3DLIGHT_POINT;
        g_light.Position.x = g_cameraPos.x;
        g_light.Position.y = g_cameraPos.y;
        g_light.Position.z = g_cameraPos.z;
        break;

    case SPOT_LIGHT:
        g_light.Type = D3DLIGHT_SPOT;
        g_light.Direction.x = 0.0f;
        g_light.Direction.y = 0.0f;
        g_light.Direction.z = 1.0f;
        g_light.Position.x = g_cameraPos.x;
        g_light.Position.y = g_cameraPos.y;
        g_light.Position.z = g_cameraPos.z;
        g_light.Theta = LIGHT_SPOT_INNER_CONE;
        g_light.Phi = LIGHT_SPOT_OUTER_CONE;
        break;
    }
}