//-----------------------------------------------------------------------------
// Copyright (c) 2006-2007 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
//
// These shaders implement Direct3D's fixed function lighting model as
// describe in:
//  DirectX Documentation for C++ > DirectX Graphics > Direct3D 9 >
//  Programming Guide > Getting Started > Lights and Materials >
//  Mathematics of Lighting.
//
// All lighting calculations are performed in eye space. The shaders only
// support a single material and light source.
//
//-----------------------------------------------------------------------------

struct Light
{
	float3 dir;             // world space direction
	float3 pos;             // world space position
	float4 ambient;
	float4 diffuse;
	float4 specular;
	float3 attenuation;     // [0]=constant [1]=linear [2]=quadratic
	float falloff;          // spot light falloff
	float theta;            // spot light inner cone
	float phi;              // spot light outer cone
};

struct Material
{
	float4 ambient;
	float4 diffuse;
	float4 emissive;
	float4 specular;
	float shininess;
};

//-----------------------------------------------------------------------------
// Globals.
//-----------------------------------------------------------------------------

float4x4 worldViewProjectionMatrix;
float4x4 worldViewMatrix;
float4x4 worldViewInverseTransposeMatrix;
float4x4 viewMatrix;

float4 globalAmbient;

Light light;
Material material;

//-----------------------------------------------------------------------------
// Textures.
//-----------------------------------------------------------------------------

texture colorMapTexture;

sampler2D colorMap = sampler_state
{
	Texture = <colorMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
};

//-----------------------------------------------------------------------------
// Vertex Shaders.
//-----------------------------------------------------------------------------

struct VS_INPUT
{
	float3 position : POSITION;
	float2 texCoord : TEXCOORD0;
	float3 normal : NORMAL;
};

struct VS_OUTPUT
{
	float4 position : POSITION;
	float2 texCoord : TEXCOORD0;
	float4 color : COLOR;
};

VS_OUTPUT VS_DirLighting(VS_INPUT IN)
{
	VS_OUTPUT OUT;

	OUT.position = mul(float4(IN.position, 1.0f), worldViewProjectionMatrix);
	OUT.texCoord = IN.texCoord;

	float4 pos = mul(float4(IN.position, 1.0f), worldViewMatrix);
	float4 n = mul(float4(IN.normal, 0.0f), worldViewInverseTransposeMatrix);
	float4 l = -normalize(mul(float4(light.dir, 0.0f), viewMatrix));
	float4 v = -normalize(float4(pos.xyz, 0.0f));
	float4 h = normalize(l + v);

	float nDotL = saturate(dot(n, l));
    float nDotH = saturate(dot(n, h));
    float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);

	float4 ambientColor = material.ambient * (globalAmbient + light.ambient);
	float4 diffuseColor = material.diffuse * light.diffuse * nDotL;
	float4 specularColor = material.specular * light.specular * power;

	OUT.color = material.emissive + ambientColor + diffuseColor + specularColor;

	return OUT;
}

VS_OUTPUT VS_PointLighting(VS_INPUT IN)
{
    VS_OUTPUT OUT;
    
    OUT.position = mul(float4(IN.position, 1.0f), worldViewProjectionMatrix);
	OUT.texCoord = IN.texCoord;
    
    float4 pos = mul(float4(IN.position, 1.0f), worldViewMatrix);
    float4 lightPos = mul(float4(light.pos, 1.0f), viewMatrix);
    float4 l = normalize(lightPos - pos);
	float d = length(l);
	float atten = 1.0f / (light.attenuation[0] +
	                     light.attenuation[1] * d +
	                     light.attenuation[2] * d * d);
    l = normalize(l);
    
    float4 n = mul(float4(IN.normal, 0.0f), worldViewInverseTransposeMatrix);
	float4 v = -normalize(float4(pos.xyz, 0.0f));
    float4 h = normalize(l + v);

	float nDotL = saturate(dot(n, l));
    float nDotH = saturate(dot(n, h));
    float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);
    
    float4 ambientColor = material.ambient * (globalAmbient + (atten * light.ambient));
	float4 diffuseColor = material.diffuse * light.diffuse * nDotL * atten;
	float4 specularColor = material.specular * light.specular * power * atten;

	OUT.color = material.emissive + ambientColor + diffuseColor + specularColor;
    
    return OUT;
}

VS_OUTPUT VS_SpotLighting(VS_INPUT IN)
{
    VS_OUTPUT OUT;
    
    OUT.position = mul(float4(IN.position, 1.0f), worldViewProjectionMatrix);
	OUT.texCoord = IN.texCoord;
    
    float4 pos = mul(float4(IN.position, 1.0f), worldViewMatrix);
    
    float4 spotPos = mul(float4(light.pos, 1.0f), viewMatrix);
    float4 spotDir = normalize(mul(float4(light.dir, 0.0f), viewMatrix));
    float4 l = spotPos - pos;
    	
	float d = length(l);
	float atten = 1.0f / (light.attenuation[0] +
	                     light.attenuation[1] * d +
	                     light.attenuation[2] * d * d);
    l = normalize(l);
    
    float spotEffect = 0.0f;
    float cosAlpha = dot(-l, spotDir);
    float cosHalfPhi = cos(light.phi * 0.5f);
    float cosHalfTheta = cos(light.theta * 0.5f);
        
    if (cosAlpha <= cosHalfPhi)
        spotEffect = 0.0f;
    else if (cosAlpha > cosHalfTheta)
        spotEffect = 1.0f;
    else
        spotEffect = pow(((cosAlpha - cosHalfPhi) / (cosHalfTheta - cosHalfPhi)), light.falloff);
       
    atten *= spotEffect;
        
    float4 n = mul(float4(IN.normal, 0.0f), worldViewInverseTransposeMatrix);
	float4 v = -normalize(float4(pos.xyz, 0.0f));
    float4 h = normalize(l + v);

	float nDotL = saturate(dot(n, l));
    float nDotH = saturate(dot(n, h));
    float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);
    
    float4 ambientColor = material.ambient * (globalAmbient + (atten * light.ambient));
	float4 diffuseColor = material.diffuse * light.diffuse * nDotL * atten;
	float4 specularColor = material.specular * light.specular * power * atten;

	OUT.color = material.emissive + ambientColor + diffuseColor + specularColor;
    
    return OUT;
}

//-----------------------------------------------------------------------------
// Pixel Shaders.
//-----------------------------------------------------------------------------

float4 PS_VertexLighting(VS_OUTPUT IN) : COLOR
{
	return IN.color * tex2D(colorMap, IN.texCoord);
}

//-----------------------------------------------------------------------------
// Techniques.
//-----------------------------------------------------------------------------

technique PerVertexDirectionalLighting
{
	pass
	{
		VertexShader = compile vs_2_0 VS_DirLighting();
		PixelShader = compile ps_2_0 PS_VertexLighting();
	}
}

technique PerVertexPointLighting
{
    pass
    {
        VertexShader = compile vs_2_0 VS_PointLighting();
        PixelShader = compile ps_2_0 PS_VertexLighting();
    }
}

technique PerVertexSpotLighting
{
    pass
    {
        VertexShader = compile vs_2_0 VS_SpotLighting();
        PixelShader = compile ps_2_0 PS_VertexLighting();
    }
}
